/* Copyright (c) 2015 Douglas Henke <dhenke@mythopoeic.org>
   This is free software -- see COPYING for details.           */
#include <stdint.h>
#include <avr/pgmspace.h>
#include <string.h>
#include <uWireM/uWireM.h>
#include "alpha15x4.h"

static const uint16_t tbl_ltr[] PROGMEM = {
   /*0PNMLKJHGgFEDCBA -- segment bits; G=G2, g=g1, P=decimal, 0=don't care */
   0b0000000011110111, /* A */
   0b0001001010001111, /* B */
   0b0000000000111001, /* C */
   0b0001001000001111, /* D */
   0b0000000001111001, /* E */
   0b0000000001110001, /* F */
   0b0000000010111101, /* G */
   0b0000000011110110, /* H */
   0b0001001000001001, /* I */
   0b0000000000011110, /* J */
   0b0010010001110000, /* K */
   0b0000000000111000, /* L */
   0b0000010100110110, /* M */
   0b0010000100110110, /* N */
   0b0000000000111111, /* O */
   0b0000000011110011, /* P */
   0b0010000000111111, /* Q */
   0b0010000011110011, /* R */
   0b0000000011101101, /* S -- FIXME: can't distinguish from 5 */
   0b0001001000000001, /* T */
   0b0000000000111110, /* U */
   0b0000101000110000, /* V */
   0b0010100000110110, /* W */
   0b0010110100000000, /* X */
   0b0001010100000000, /* Y */
   0b0000110000001001  /* Z */
};

static const uint16_t tbl_num[] PROGMEM = {
   /*0PNMLKJHGgFEDCBA -- segment bits; G=G2, g=g1, P=decimal, 0=don't care */
   0b0000110000111111, /* 0 */
   0b0001001000000000, /* 1 */
   0b0000000011011011, /* 2 */
   0b0000000011001111, /* 3 */
   0b0000000011100110, /* 4 */
   0b0000000011101101, /* 5 -- FIXME: can't distinguish from S */
   0b0000000011111101, /* 6 */
   0b0000000000000111, /* 7 */
   0b0000000011111111, /* 8 */
   0b0000000011101111, /* 9 */
};

#define SEG_STAR 0b0011111111000000 /* "*" (used for undefined chars) */

/*** alpha15x4_cmd() -- send command to LED driver

Sends a single-byte command to the LED driver. Examples of such commands
would be those to set the brightness or blink mode.

The argument is the command value to be sent.
***/
void alpha15x4_cmd(const uint8_t x) {
   uint8_t buf[2];
   buf[0] = uWireM_addr_send(ALPHA15X4_I2C_ADDR);
   buf[1] = x;
   uWireM_xfer(buf, 2);
}

/*** alpha15x4_clear() -- turn all segments off

Blanks the LED display by writing all-segments-off data.
***/
void alpha15x4_clear(void) {
   uint8_t buf[10];
   /* Why bother clearing buf[0] if we're going to immediately set it
      to another value? Avoiding doing so would make the code bigger (and
      probably slower, too). */
   memset(buf, 0, sizeof(buf));
   buf[0] = uWireM_addr_send(ALPHA15X4_I2C_ADDR);
   /* buf[1] = 0x00 -- start writing at address 0x00 */
   uWireM_xfer(buf, 10);
}

/*** alpha15x4_draw() -- draw a letter or digit on one display position

Writes character data for a single character into one of the four display
positions on the 14-segment LED array.

Arguments:
   pos -- character position to update, from 0 on the left to 3 on the right
   c -- ASCII character to draw
   dp -- if non-0, turn on decimal point in this position

Notes:
   o- alpha15x4_init() must be called before calling this function
   o- Only upper-case alpha characters A-Z, digits 0-9 and space are supported.
      Lower-case alpha characters a-z are folder to their upper-case
      equivalents. Any other character value renders as a star (*).
***/
void alpha15x4_draw(const uint8_t pos, const char c, const uint8_t dp) {
   uint8_t buf[4];
   uint16_t *p = (uint16_t *)(buf+2);

   if(c >= 'A' && c <= 'Z')       /* upper-case alpha */
      *p = pgm_read_word(tbl_ltr + c - 'A');
   else if(c >= 'a' && c <= 'z')  /* lower-case alpha (folded to upper) */
      *p = pgm_read_word(tbl_ltr + c - 'a');
   else if(c >= '0' && c <= '9')  /* digit 0-9 */
      *p = pgm_read_word(tbl_num + c - '0');
   else if(c == ' ') *p = 0x0000; /* space */
   else *p = SEG_STAR; /* any other character renders as a star (*) */

   if(dp) *p |= 0b0100000000000000; /* turn on decimal point if requested */

   buf[0] = uWireM_addr_send(ALPHA15X4_I2C_ADDR);
   buf[1] = pos << 1; /* start writing at 0x00 + 2*pos */
   uWireM_xfer(buf, 4);
}
