/* Copyright (c) 2015 Douglas Henke <dhenke@mythopoeic.org>
   This is free software -- see COPYING for details.           */
#include <string.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "rotary_encoder.h"

/* Assumption: All three rotary encoder inputs are on pins which share
   the same common DDR, PORT and PIN register. */
#define RENC_DDR DDRB
#define RENC_PORT PORTB
#define RENC_PIN PINB

/* Register and bit definitions for the button press: */
#define RENC_BTN_DDRBIT  _BV(DDB1)
#define RENC_BTN_PORTBIT _BV(PORTB1)
#define RENC_BTN_PINBIT  _BV(PINB1)
#define RENC_BTN_INTBIT  _BV(PCINT1)

/* Register and bit definitions for the first of two rotation switches: */
#define RENC_A_DDRBIT    _BV(DDB3)
#define RENC_A_PORTBIT   _BV(PORTB3)
#define RENC_A_PINBIT    _BV(PINB3)
#define RENC_A_INTBIT    _BV(PCINT3)

/* ... and the second of two rotation switches: */
#define RENC_B_DDRBIT    _BV(DDB4)
#define RENC_B_PORTBIT   _BV(PORTB4)
#define RENC_B_PINBIT    _BV(PINB4)
#define RENC_B_INTBIT    _BV(PCINT4)

#define RENC_ROT_PINBITS (RENC_A_PINBIT | RENC_B_PINBIT)
#define RENC_BTN (RENC_PIN & RENC_BTN_PINBIT) /* zero IFF pressed */
#define RENC_A (RENC_PIN & RENC_A_PINBIT)
#define RENC_B (RENC_PIN & RENC_B_PINBIT)

/* The rotary encoder acts like three simple switches, each of which
   connect a pin either directly to ground or to nothing. Therefore,
   all three need a pull-up. However PINB3 on the Trinket has an
   external 1.5K pull-up resistor connected, so we don't want to use
   the internal pull-up on the ATTiny85 for it. */
#define RENC_PULLUP (RENC_BTN_PORTBIT | RENC_B_PORTBIT)

volatile evt_t _evt = { BTN_NONE, 0 };

static volatile uint8_t _prev_renc = 0; /* previous rotary encoder state */
static volatile uint8_t _prev_btn = 0; /* previous button state */

/* Since we need to measure intervals longer than what we can get from
   an 8-bit timer using the slowest available clock, we have an additional
   counter (_timer_countdown) which is decremented each time the timer
   interrupt handler is called. The interval we're timing has expired when
   the interrupt handler is called with _timer_countdown equal to zero. */
static volatile uint8_t _timer_countdown = 0;

/* _timer is the state of the button handling timer. TS_NONE means the
   timer isn't running. TS_DEBOUNCE means that the button has just changed
   state, and we are ignoring further state changes until the timer expires
   (since they are likely due to contact noise rather than deliberate user
   action). TS_LONGPRESS means the button is being held down, and we will
   generate a BTN_LONG event when the timer expires (unless the button is
   released before then). */
static volatile enum __attribute__((__packed__)) {
    TS_NONE=0, TS_DEBOUNCE, TS_LONGPRESS
} _timer = TS_NONE;

/* The timer ticks at F_CPU/1024, or 8MHz/1024 = 7812.5Hz. Thus, each tick is
   1000 ms/s / 7812.5Hz = 0.128ms. For debouncing, we want about 50ms, which
   is 390.625 ticks. The timer is only 8 bits, so we have to let it fire
   multiple times and use an external counter. A solution with the fewest
   interrupts is preferrable, so we set the timer for 195 ticks and let it
   expire twice. That yields 49.92ms, which is plenty close (since 50ms is
   just a number I made up that seems reasonable). */
#define TIMER_DEBOUNCE_TICKS 195
#define TIMER_DEBOUNCE_COUNTDOWN 1 /* zero means one pass */

/* Likewise, the long-press interval we want is about 800ms, or 6250 ticks.
An exact fewest-interrupts solution is 25 passes of 250 ticks. */
#define TIMER_LONGPRESS_TICKS 250
#define TIMER_LONGPRESS_COUNTDOWN 25

/*** ISR(PCINT0_vect) -- pin-change interrupt handler

Called when any of the three rotary encoder inputs (A, B, push button) change
state.
***/
ISR(PCINT0_vect) {
   static uint8_t prevdir = 1; /* rotary encoder rotation direction at the
                                  time of the last good read */
   uint8_t x; /* hold input value snapshot (so it doesn't change while
                 we're trying to work with it) */

   /* rotary encoder handling: */
   x = PINB & RENC_ROT_PINBITS;
   if(_prev_renc != x) { /* if either of the two inputs A or B changed... */

      /* If both bits changed, we missed a pulse. */
      if((_prev_renc ^ x) == RENC_ROT_PINBITS) {
         /* Assume we went two state changes in whatever direction we saw
            last. */
         _evt.rotary += (prevdir << 1);
      }

      /* If bit A of this sample is the same as bit B of the
         previous sample, this is clockwise rotation. */
      else if(((_prev_renc & RENC_B_PINBIT)?RENC_A_PINBIT:0) ==
       (x & RENC_A_PINBIT)) {
         _evt.rotary++; prevdir = 1;
      }

      /* Else, bit A of this sample is different from bit B of
         the previous sample, and this is counterclockwise. */
      else { _evt.rotary--; prevdir = -1; }

      _prev_renc = x;
   }

   /* push button handling */
   x = RENC_BTN;
   if(_prev_btn != x) {

      switch(_timer) {
         case TS_LONGPRESS:
            /* stop the timer */
            TCCR0B &= ~(_BV(CS02) | _BV(CS01) | _BV(CS00)); /* stop timer */

            /* The following is subtle: It is possible that the longpress
               timer expired between when we entered the PCINT0 ISR and here.
               If we left things alone, the timer interrupt handler would
               get called the moment we returned from this pin-change handler,
               and it (the timer interrupt handler) would have nothing to do
               and be confused and sad. */
            TIFR |= OCF0A; /* clear pending timer interrupt (if any) */
            _timer = TS_NONE;
            /* fall-through to TS_NONE intentional */

         case TS_NONE:
            _evt.btn = x ? BTN_UP : BTN_DOWN;
            _prev_btn = x;

            /* start the debounce timer: */
            TCNT0 = 0; /* start timer counter at zero */
            OCR0A = TIMER_DEBOUNCE_TICKS; /* expire when timer reaches this */
            TCCR0B |= _BV(CS02) | _BV(CS00); /* clock prescaler: 1/1024 */
            _timer = TS_DEBOUNCE;
            _timer_countdown = TIMER_DEBOUNCE_COUNTDOWN;
            break;

         case TS_DEBOUNCE: break; /* ignore changes while debounce timer runs */
      }
   }
}

/*** ISR(TIM0_COMPA_vect) -- timer 0 compare match A interrupt handler

Called when timer 0 value matches OCR0A register. We use timer0 for
debouncing the push button, and for detecting "long press" events (where the
user presses the button and holds it down).
***/
ISR(TIM0_COMPA_vect) {
   uint8_t x = RENC_BTN;

   /* If _timer_countdown is non-zero, just decrement it and return. Only
      if this interrupt handler is called with _timer_countdown zero do we
      go on to do anything else. This effectively lets us use _timer_countdown
      as a multiplier to get longer intervals than we could get with an
      8-bit timer alone. */
   if(_timer_countdown) { _timer_countdown--; return; }

   switch(_timer) { /* Do different things depending on why the timer was
                       running in the first place: */
      case TS_NONE: break; /* "never happens" */

      case TS_DEBOUNCE:
         if(x != _prev_btn) _evt.btn = x ? BTN_UP : BTN_DOWN;

         if(!x) { /* button pressed */
            /* start the longpress timer: */
            TCNT0 = 0; /* start timer counter at zero */
            OCR0A = TIMER_LONGPRESS_TICKS; /* expire when timer reaches this */
            TCCR0B |= _BV(CS02) | _BV(CS00); /* clock prescaler: 1/1024 */
            _timer = TS_LONGPRESS;
            _timer_countdown = TIMER_LONGPRESS_COUNTDOWN;

         } else { /* button not pressed */
            TCCR0B &= ~(_BV(CS02) | _BV(CS01) | _BV(CS00)); /* stop timer */
            _timer = TS_NONE;
         }
         break;

      case TS_LONGPRESS:
         if(!x) _evt.btn = BTN_LONG;
         TCCR0B &= ~(_BV(CS02) | _BV(CS01) | _BV(CS00)); /* stop timer */
         _timer = TS_NONE;
         break;
   }
}

/*** renc_setup(void) -- initialize rotary encoder inputs and timer

This function must be called before input events will be received.
Once renc_setup() returns, the (volatile) global _evt structure will
change to reflect user input. (Note: code which examines and/or modifies
_evt must run with interrupts disabled!)

This function does not enable global interrupts (via sei() or equivalent) --
that is left to the caller, as there may be other initialization to be done
before interrupts are wanted.
***/
void renc_setup(void) {
   /* Set rotary encoder pins as input: */
   /* (Nothing to do; all pins default to input on reset...) */
   /* RENC_DDR &= ~(RENC_BTN_DDRBIT | RENC_A_DDRBIT | RENC_B_DDRBIT); */

   /* activate pull-ups on rotary encoder inputs as needed: */
   RENC_PORT |= RENC_PULLUP;

   /* Select "CTC" (clear timer on compare) mode by setting the WGM
      (waveform generation) bits { WGM00, WGM01, WGM02 } to { 0, 1, 0 }.
      Since these bits default to zero, we just have to turn on WGM01: */
   TCCR0A |= _BV(WGM01);
   TIMSK |= _BV(OCIE0A); /* enable interrupts on timer 0 compare match A */

   _prev_renc = PINB & (RENC_A_PINBIT | RENC_B_PINBIT);
   _prev_btn = RENC_BTN;

   GIMSK |= (1 << PCIE); /* enable pin-change interrupts */
   /* send interrupt for rotary encoder and button press changes */
   PCMSK |= RENC_BTN_INTBIT | RENC_A_INTBIT | RENC_B_INTBIT;
}
